#include <math.h>

#include <qlayout.h>
#include <qpainter.h>

#include <netwm.h>
#undef Bool // to avoid problem in qvariant.h

#include <kapp.h>
#include <kstddirs.h>
#include <kconfig.h>
#include <kwin/workspace.h>
#include <kwin/options.h>
#include <kpixmap.h>
#include <kpixmapeffect.h>
#include <kdrawutil.h>

#include "logik.h"
#include "logikbuttons.h"

using namespace KWinInternal;

const uint toolSideMargin     = 3;
const uint normalSideMargin   = 5;
const uint toolBottomMargin   = 4;
const uint normalBottomMargin = 6;
const uint normalTextVMargin  = 4;
const uint toolTextVMargin    = 1;

bool      doInit                = true;
uint      toolTitleHeight;
uint      normalTitleHeight;
KPixmap * toolBgGradient      = 0;
KPixmap * toolALeftEdge       = 0;
KPixmap * toolPLeftEdge       = 0;
KPixmap * normalBgGradient    = 0;
KPixmap * normalALeftEdge     = 0;
KPixmap * normalPLeftEdge     = 0;

extern "C"
{
  Client * allocate(Workspace * ws, WId w, int tool)
  {
    return new Logik(ws, w, tool != 0);
  } /* allocate */
  void init()
  {
    // Nothing... Init needs Client API -> done in first client.
    doInit = true;
  } /* init */
  void reset()
  {
    // Indicates that elements are to be recalculated.
    doInit = true;
  } /* reset */
  void deinit()
  {
    delete toolBgGradient;
    delete toolALeftEdge;
    delete toolPLeftEdge;
    delete normalBgGradient;
    delete normalALeftEdge;
    delete normalPLeftEdge;
  } /* deinit */
}

Logik::Logik(Workspace * ws, WId w, bool tool, QWidget * parent, const char * name)
        : Client(ws, w, parent, name, WResizeNoErase),
          tool(tool),
          mainLayout(0),
          middleSpacer(0),
          showStayOnTop(true)
{
  setBackgroundMode(NoBackground);

  resetLayout();

  leftButtonList.setAutoDelete(true);
  rightButtonList.setAutoDelete(true);

  connect(options, SIGNAL(resetClients()), this, SLOT(slotReset()));
} // Logik::Logik

Logik::~Logik()
{

} // Logik::~Logik

void Logik::resizeEvent(QResizeEvent * e)
{
  Client::resizeEvent(e);
  doShape();
  repaint();
} // Logik::resizeEvent

void Logik::captionChange(const QString &)
{
  repaint();
} // Logik::captionChange

void Logik::paintEvent(QPaintEvent * pe)
{
  QRect titleRect(middleSpacer->geometry());
  titleRect.setTop(1);
  QFontMetrics fm(options->font(isActive(), tool));
  QPixmap* bgpixmap = (tool?toolBgGradient:normalBgGradient);
  QPixmap* leftedge = (tool?(isActive()?toolALeftEdge:toolPLeftEdge):
                            (isActive()?normalALeftEdge:normalPLeftEdge));

  int r(width());
  int b(height());
  int ht(titleRect.height());
  int mt(ht/2+1);
  int delta(mt-(tool?1:3));
  int dec(titleRect.x());
  int capw(fm.width(caption()));
  int ch(ht+1-(tool?2:3));
  int ew(leftedge->width());

  if (ew*2 > titleRect.width())
    capw = 0;
  else
    capw = QMIN(capw, titleRect.width()-2*ew);

  QPainter p(this);

  if (!capw)
    p.drawTiledPixmap(2, 2, width()-4, ht-2, *bgpixmap);
  else
  {
    KPixmap * cappixmap = new KPixmap();
    cappixmap->resize(capw, ch);
    QPainter p2(cappixmap);
    p2.setPen(options->color(Options::TitleBar, isActive()));
    p2.setBrush(options->color(Options::TitleBar, isActive()));
    p2.drawRect(0, (tool?0:1), capw, ch-(tool?0:2));
    if (tool)
    {
      p2.setPen(colorGroup().background().dark());
      p2.drawLine(0, ch-1, capw-1, ch-1);
    }
    else
    {
      p2.setPen(colorGroup().background().dark());
      p2.drawLine(0, 0, capw-1, 0);
      p2.setPen(colorGroup().background().light());
      p2.drawLine(0, ch-2, capw-1, ch-2);
      p2.setPen(colorGroup().background().dark(110));
      p2.drawLine(0, ch-1, capw-1, ch-1);
    }
    p2.setFont(options->font(isActive(), tool));
    p2.setPen(options->color(Options::Font, isActive()));
    p2.drawText(0, 1, capw, ch-3, AlignVCenter, caption());
    p2.end();
    p.drawTiledPixmap(2, 2, dec, ht-2, *bgpixmap);
    p.drawPixmap(dec+2, 2, *leftedge);
    p.drawPixmap(dec+2+ew, 2, *cappixmap);
    p.drawTiledPixmap(dec+2+2*ew+capw, 2, r-dec-4-capw-2*ew, ht-2, *bgpixmap);
    p.flush();
    p.save();
    p.setWorldMatrix(QWMatrix(-1, 0, 0, 1, r, 0), true);
    p.drawPixmap(r-dec-1-2*ew-capw, 2, *leftedge);
    p.flush();
    p.restore();
    delete cappixmap;
  }

  p.setPen(colorGroup().background());
  p.setBrush(colorGroup().background());
  p.drawRect(1, ht, r-4, b-ht-3);
  p.drawRect(r-4, ht, r-3, b-3);
  p.drawLine(4, b-5, r-5, b-5);
  p.drawLine(4, b-4, 16, b-4);
  p.drawLine(4, b-3, 14, b-3);
  p.drawLine(r-5, b-4, r-17, b-4);
  p.drawLine(r-5, b-3, r-15, b-3);
  p.setPen(options->color(Options::TitleBar, isActive()));
  p.drawLine(17, b-2, r-18, b-2);
  p.drawLine(19, b-3, r-20, b-3);
  p.setPen(colorGroup().background().dark());
  p.drawLine(4, ht, r-5, ht);
  p.drawLine(4, ht, 4, b-6);
  p.drawLine(r-2, 2, r-2, b-3);
  p.drawLine(2, b-2, 17, b-2);
  p.drawLine(r-18, b-2, r-3, b-2);
  p.drawLine(15, b-3, 19, b-3);
  p.drawLine(r-20, b-3, r-16, b-3);
  p.drawLine(17, b-4, r-18, b-4);
  p.setPen(colorGroup().background().light());
  p.drawLine(2, 1, r-3, 1);
  p.drawLine(1, 2, 1, b-3);
  p.drawLine(r-5, ht, r-5, b-6);
  p.drawLine(4, b-6, r-5, b-6);

  p.setPen(Qt::black);
  p.setBrush(Qt::NoBrush);
  p.drawRect(rect());
  p.drawPoint(1, 1);
  p.drawPoint(1, b-2);
  p.drawPoint(r-2, 1);
  p.drawPoint(r-2, b-2);
  p.flush();
}

void Logik::doShape()
{
  QRegion mask(0, 0, width(), height());

  int r(width());
  int b(height());

  mask -= QRegion(0, 0, 1, 2);
  mask -= QRegion(r - 1, 0, 1, 2);
  mask -= QRegion(0, b - 2, 1, 2);
  mask -= QRegion(r - 1, b - 2, 1, 2);
  mask -= QRegion(1, 0, 1, 1);
  mask -= QRegion(r - 2, 0, 1, 1);
  mask -= QRegion(1, b - 1, 1, 1);
  mask -= QRegion(r - 2, b - 1, 1, 1);

  setMask(mask);
} // Logik::doShape

void Logik::showEvent(QShowEvent *)
{
  doShape();
  repaint();
} // Logik::showEvent

void Logik::windowWrapperShowEvent(QShowEvent *)
{
  doShape();
  repaint();
} // Logik::windowWrapperShowEvent

void Logik::mouseDoubleClickEvent(QMouseEvent * e)
{
  if (middleSpacer->geometry().contains(e->pos()))
    workspace()->performWindowOperation(this, options->operationTitlebarDblClick());
} // Logik::mouseDoubleClickEvent

void Logik::stickyChange(bool b)
{
  emit(stkyChange(b));
} // Logik::stickyChange

void Logik::maximizeChange(bool b)
{
  emit(maxChange(b));
} // Logik::maximizeChange

void Logik::activeChange(bool)
{
  repaint();
} // Logik::activeChange

Client::MousePosition Logik::mousePosition(const QPoint & p) const
{
  int x = p.x();
  int y = p.y();

  uint titleBarHeight = middleSpacer->geometry().height();
  if (y < titleBarHeight)
    if ((x > 4) && (x < width() - 5))
      if (y < 3)
        return Client::Top;
      else
        return Client::Center;
    else
      if (x < titleBarHeight+1)
        return Client::TopLeft;
      else
        return Client::TopRight;
  else if (y < height() - 20)
    if (x < 4)
      return Client::Left;
    else
      return Client::Right;
  else
    if (x < 20)
      return Client::BottomLeft;
    else
      if (x > width() - 20)
        return Client::BottomRight;
      else
        return Client::Bottom;

  return Client::mousePosition(p);
} // Logik::mousePosition

void Logik::slotReset()
{
  resetLayout();
  repaint();
} // Logik::slotReset

void Logik::slotMaximize(int button)
{
  switch (button)
  {
    case MidButton:
      maximize(MaximizeVertical);
      break;

    case RightButton:
      maximize(MaximizeHorizontal);
      break;

    case LeftButton:
    default:
      maximize(MaximizeFull);
  }
} // Logik::slotMaximize

void Logik::toggleStayOnTop()
{
  setStaysOnTop(!staysOnTop());
} // Logik::toggleStayOnTop

LogikButton * Logik::createButton(const QString & s, QWidget * parent)
{
  LogikButton * b = 0;

  if (("Help" == s) && providesContextHelp())
  {
    b = new LogikButtonHelp(parent, "Help");
    connect(b, SIGNAL(help()), this, SLOT(contextHelp()));
  }

  else if ("Sticky" == s)
  {
    b = new LogikButtonSticky(isSticky(), parent, "Sticky");
    connect(b, SIGNAL(toggleSticky()), this, SLOT(toggleSticky()));
    connect(this, SIGNAL(stkyChange(bool)), b, SLOT(slotStickyChange(bool)));
  }

  else if ("StayOnTop" == s)
  {
    b = new LogikButtonStayOnTop(parent, "Stay on top");
    connect(b, SIGNAL(toggleStayOnTop()), this, SLOT(toggleStayOnTop()));
  }

  else if ("Iconify" == s && isMinimizable())
  {
    b = new LogikButtonIconify(parent, "Minimize");
    connect(b, SIGNAL(iconify()), this, SLOT(iconify()));
  }

  else if ("Maximize" == s && isMaximizable())
  {
    b = new LogikButtonMaximize(isMaximized(), parent, "Maximize");
    connect(b, SIGNAL(maximize(int)), this, SLOT(slotMaximize(int)));
    connect(this, SIGNAL(maxChange(bool)), b, SLOT(slotMaximizeChange(bool)));
  }

  else if ("Close" == s)
  {
    b = new LogikButtonClose(parent, "Close");
    connect(b, SIGNAL(closeWindow()), this, SLOT(closeWindow()));
  }

  else if ("Lower" == s)
  {
    b = new LogikButtonLower(parent, "Lower");
    connect(b, SIGNAL(lowerWindow()), this, SLOT(slotLowerWindow()));
  }

  return b;
}

void Logik::createButtons()
{
  leftButtonList.clear();
  rightButtonList.clear();

  QString buttons = options->titleButtonsLeft() + "|T" + options->titleButtonsRight();
  QPtrList<LogikButton> *buttonList = &leftButtonList;
  for (unsigned int i = 0; i < buttons.length(); ++i)
  {
    LogikButton * tb = 0;
    switch (buttons[i].latin1())
    {
      case 'S': // Sticky
        tb = createButton("Sticky", this);
        break;

      case 'H': // Help
        tb = createButton("Help", this);
        break;

      case 'I': // Minimize
        tb = createButton("Iconify", this);
        break;

      case 'A': // Maximize
        tb = createButton("Maximize", this);
        break;

      case 'X': // Close
        tb = createButton("Close", this);
        break;

      case 'T': // Stay On Top
        tb = createButton("StayOnTop", this);
        break;

      case '|':
        buttonList = &rightButtonList;
        break;
    }
    if (0 != tb)
      buttonList->append(tb);
  }
} // Logik::createButtons

void Logik::calculateElements()
{
  doInit = false;

    // Deleting old pixmaps (if any).
  if (toolBgGradient) delete toolBgGradient;
  if (toolALeftEdge) delete toolALeftEdge;
  if (toolPLeftEdge) delete toolPLeftEdge;
  if (normalBgGradient) delete normalBgGradient;
  if (normalALeftEdge) delete normalALeftEdge;
  if (normalPLeftEdge) delete normalPLeftEdge;

  QFontMetrics fm(options->font(isActive(), false));
  QFontMetrics fmt(options->font(isActive(), true));

  normalTitleHeight = QMAX(14, fm.height() + normalTextVMargin * 2);
  if (0 != normalTitleHeight % 2)
    normalTitleHeight += 1;
  toolTitleHeight = QMAX(14, fmt.height() + toolTextVMargin * 2);
  if (0 != toolTitleHeight % 2)
    toolTitleHeight += 1;

  int m;
  QPointArray a;

    // Tool box pixmaps
  toolBgGradient = new KPixmap();
  toolBgGradient->resize(32, toolTitleHeight-3);
  KPixmapEffect::gradient(*toolBgGradient,
                          colorGroup().background().light(110),
                          colorGroup().background().dark(110),
                          KPixmapEffect::VerticalGradient);
  m = toolBgGradient->height()/2 - 1;
  QPainter *p = new QPainter(toolBgGradient);
  p->setPen(colorGroup().background().dark());
  p->drawLine(0, m, 31, m);
  p->setPen(colorGroup().background().light());
  p->drawLine(0, m+1, 31, m+1);
  delete p;
  toolALeftEdge = new KPixmap();
  toolALeftEdge->resize(m+3, toolTitleHeight-3);
  p = new QPainter(toolALeftEdge);
  p->drawPixmap(0, 0, *toolBgGradient);
  p->setPen(colorGroup().background().dark());
  p->drawLine(1, m, m+1, 0);
  p->setPen(colorGroup().background().light());
  p->drawLine(1, m+1, m+1, m*2+1);
  p->setPen(options->color(Options::TitleBar, true));
  p->setBrush(options->color(Options::TitleBar, true));
  a.setPoints(4, 2, m, m+2, 0, m+2, 2*m+1, 2, m+1);
  p->drawPolygon(a);
  delete p;
  toolPLeftEdge = new KPixmap(*toolALeftEdge);
  p = new QPainter(toolPLeftEdge);
  p->setPen(options->color(Options::TitleBar, false));
  p->setBrush(options->color(Options::TitleBar, false));
  p->drawPolygon(a);
  delete p;

    // Normal window pixmaps
  normalBgGradient = new KPixmap();
  normalBgGradient->resize(32, normalTitleHeight-3);
  KPixmapEffect::gradient(*normalBgGradient,
                          colorGroup().background().light(110),
                          colorGroup().background().dark(110),
                          KPixmapEffect::VerticalGradient);
  m = normalBgGradient->height()/2 - 1;
  p = new QPainter(normalBgGradient);
  p->setPen(colorGroup().background().dark());
  p->drawLine(0, m, 31, m);
  p->setPen(colorGroup().background().light());
  p->drawLine(0, m+1, 31, m+1);
  delete p;
  normalALeftEdge = new KPixmap();
  normalALeftEdge->resize(m+3, normalTitleHeight-3);
  p = new QPainter(normalALeftEdge);
  p->drawPixmap(0, 0, *normalBgGradient);
  p->setPen(options->color(Options::TitleBar, true));
  p->setBrush(options->color(Options::TitleBar, true));
  a.setPoints(6, 2, m, m+1, 1, m+2, 1, m+2, 2*m, m+1, 2*m, 2, m+1);
  p->drawPolygon(a);
  p->setPen(colorGroup().background().dark());
  p->drawLine(1, m, m+1, 0);
  p->drawPoint(m+2, 0);
  p->setPen(colorGroup().background().light());
  p->drawLine(1, m+1, m+1, m*2+1);
  p->drawPoint(m+2, m*2+1);
  delete p;
  normalPLeftEdge = new KPixmap(*normalALeftEdge);
  p = new QPainter(normalPLeftEdge);
  p->setPen(options->color(Options::TitleBar, false));
  p->setBrush(options->color(Options::TitleBar, false));
  p->drawPolygon(a);
  delete p;

} // Logik::calculateElements

void Logik::resetLayout()
{
  if (doInit)
    calculateElements();

  const uint sideMargin       = (tool?toolSideMargin:normalSideMargin);
  const uint bottomMargin     = (tool?toolBottomMargin:normalBottomMargin);

  uint titleHeight = (tool?toolTitleHeight:normalTitleHeight);
  uint buttonWidth = titleHeight - (tool?toolTextVMargin:normalTextVMargin)*2;

  middleSpacer = new QSpacerItem(0, titleHeight, QSizePolicy::Expanding, QSizePolicy::Fixed);

  delete mainLayout;
  mainLayout = new QVBoxLayout(this, 0, 0);

  QHBoxLayout * topLayout = new QHBoxLayout(mainLayout, 0, 0);
  topLayout->setMargin(0);
  topLayout->setSpacing(4);

  createButtons();

  topLayout->addItem(new QSpacerItem(6, titleHeight, QSizePolicy::Fixed, QSizePolicy::Fixed));
  for (QListIterator<LogikButton> it(leftButtonList); it.current(); ++it)
  {
    topLayout->addWidget(it.current(), Qt::AlignVCenter);
    topLayout->setStretchFactor(it.current(), 0);
    it.current()->setFixedSize(buttonWidth, buttonWidth);
  }
  topLayout->addItem(middleSpacer);
  for (QListIterator<LogikButton> it(rightButtonList); it.current(); ++it)
  {
    topLayout->addWidget(it.current(), Qt::AlignVCenter);
    it.current()->setFixedSize(buttonWidth, buttonWidth);
  }
  topLayout->addItem(new QSpacerItem(6, titleHeight, QSizePolicy::Fixed, QSizePolicy::Fixed));

  QHBoxLayout * midLayout   = new QHBoxLayout(mainLayout, 0, 0);

  midLayout->addSpacing(sideMargin);
  midLayout->addWidget(windowWrapper());
  midLayout->addSpacing(sideMargin);

  mainLayout->addSpacing(bottomMargin);

  mainLayout->setStretchFactor(topLayout, 0);
  mainLayout->setStretchFactor(midLayout, 1);
} // Logik::resetLayout

void Logik::animateIconifyOrDeiconify(bool /* iconify */)
{
} // Logik::animateIconifyOrDeiconify
