#include <qpainter.h>

#include "logikbuttons.h"
#include <kpixmap.h>

using namespace KWinInternal;

LogikButton::LogikButton(QWidget * parent, const QString& tip)
  : KWinWidgetButton(parent, 0, 0, tip),
    mouseOver  (false),
    mouseDown  (false)
{
  setSizePolicy(QSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum));
  setBackgroundMode(NoBackground);
} // LogikButton::LogikButton

LogikButton::~LogikButton()
{
  // Empty.
} // LogikButton::~LogikButton

void LogikButton::mousePressEvent(QMouseEvent *e)
{
  KWinWidgetButton::mousePressEvent(e);
  mouseDown = true;
  repaint();
} // LogikButton::mousePressEvent

void LogikButton::mouseReleaseEvent(QMouseEvent * e)
{
  mouseDown = false;
  repaint();

  if (rect().contains(e->pos()))
  {
    clickEvent(e->button());
  }
} // LogikButton::mouseReleaseEvent

void LogikButton::enterEvent(QEvent *e)
{
  KWinWidgetButton::enterEvent(e);
  mouseOver = true;
  repaint();
} // LogikButton::enterEvent

void LogikButton::leaveEvent(QEvent *e)
{
  KWinWidgetButton::leaveEvent(e);
  mouseOver = false;
  repaint();
} // LogikButton::leaveEvent

void LogikButton::paintEvent(QPaintEvent *)
{
  QBitmap mp(width(), height(), true);
  QPainter mpp(&mp);
  mpp.setPen(Qt::color1);
  mpp.setBrush(Qt::color1);
  mpp.drawEllipse(0, 0, width(), height());
  setMask(mp);

  QPen* highlightPen = 0;

  if (mouseDown)
    highlightPen = new QPen(colorGroup().light());

  else if (mouseOver)
    highlightPen = new QPen(colorGroup().highlight());

  QPainter p(this);

  if (highlightPen)
  {
    p.setPen(*highlightPen);
    p.setBrush(colorGroup().background());
    p.drawEllipse(0, 0, width(), height());
//    p.drawArc(1, 1, width()-2, height()-2, 0, 5760);
  }
  else
  {
    p.setBrush(colorGroup().background());
    p.drawEllipse(0, 0, width(), height());
    p.setPen(colorGroup().background().light());
    p.drawArc(0, 0, width(), height(), 2880, 5760);
    p.setPen(colorGroup().background().dark());
    p.drawArc(0, 0, width(), height(), 0, 2880);
  }
  delete highlightPen;

  // Draw icon.

  QPoint center(rect().center());

  int bwby2(bitmap.width() / 2);
  int bhby2(bitmap.height() / 2);

  p.setBrush(NoBrush);
  p.setPen(Qt::black);

  p.drawPixmap(center.x() - bwby2 + 1, center.y() - bhby2 + 1, bitmap);
} // LogikButton::paintEvent

QSize LogikButton::sizeHint() const
{
  return QSize(16, 16);
} // LogikButton::sizeHint

QSize LogikButton::minimumSizeHint() const
{
  return QSize(14, 14);
} // LogikButton::minimumSizeHint

void LogikButton::setBitmap(const QBitmap & b)
{
  bitmap = b;
  repaint();
}

void LogikButton::resizeEvent(QResizeEvent *)
{
  repaint();
} // LogikButton::resizeEvent

static unsigned char close_bits[] =
{
   0x42, 0xe7, 0x7e, 0x3c, 0x3c, 0x7e, 0xe7, 0x42
};

LogikButtonClose::LogikButtonClose(QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, close_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonClose::LogikButtonClose

void LogikButtonClose::clickEvent(int)
{
  emit(closeWindow());
} // LogikButtonClose::clickEvent

static unsigned char help_bits[] =
{
   0x18, 0x18, 0x00, 0x1c, 0x18, 0x18, 0x18, 0x3c
};

LogikButtonHelp::LogikButtonHelp(QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, help_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonHelp::LogikButtonHelp

void LogikButtonHelp::clickEvent(int)
{
  emit(help());
} // LogikButtonHelp::clickEvent

static unsigned char iconify_bits[] =
{
   0x00, 0x00, 0x00, 0x7e, 0x7e, 0x3c, 0x18, 0x00
};

LogikButtonIconify::LogikButtonIconify(QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, iconify_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonIconify::LogikButtonIconify

void LogikButtonIconify::clickEvent(int)
{
  emit(iconify());
} // LogikButtonIconify::clickEvent

static unsigned char lower_bits[] =
{
   0x1f, 0x1f, 0x1f, 0xff, 0x8f, 0x88, 0x88, 0xf8
};

LogikButtonLower::LogikButtonLower(QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, lower_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonLower::LogikButtonLower

void LogikButtonLower::clickEvent(int)
{
  emit(lowerWindow());
} // LogikButtonLower::clickEvent

static unsigned char maximize_bits[] =
{
   0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80, 0x00, 0x00
};

static unsigned char unmaximize_bits[] =
{
   0x00, 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f
};

LogikButtonMaximize::LogikButtonMaximize(bool max, QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, max ? unmaximize_bits : maximize_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonMaximize::LogikButtonMaximize

void LogikButtonMaximize::slotMaximizeChange(bool max)
{
  QBitmap b(8, 8, max ? unmaximize_bits : maximize_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonMaximize::slotMaximizeChange

void LogikButtonMaximize::clickEvent(int button)
{
  emit(maximize(button));
} // LogikButtonMaximize::clickEvent

static unsigned char sticky_bits[] =
{
   0x20, 0x70, 0xfa, 0x7e, 0x3c, 0x1c, 0x32, 0x01
};

static unsigned char unsticky_bits[] =
{
   0x1c, 0x1c, 0x1c, 0x3e, 0x7f, 0x08, 0x08, 0x08
};

LogikButtonSticky::LogikButtonSticky(bool sticky, QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, sticky ? unsticky_bits : sticky_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonSticky::LogikButtonSticky

void LogikButtonSticky::slotStickyChange(bool sticky)
{
  QBitmap b(8, 8, sticky ? unsticky_bits : sticky_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonSticky::slotStickyChange

void LogikButtonSticky::clickEvent(int)
{
  emit(toggleSticky());
} // LogikButtonSticky::clickEvent

static unsigned char stayontop_bits[] =
{
   0xf8, 0x88, 0x8f, 0x8f, 0xff, 0x3f, 0x3f, 0x3f
};

LogikButtonStayOnTop::LogikButtonStayOnTop(QWidget * parent, const QString& tip)
  : LogikButton(parent, tip)
{
  QBitmap b(8, 8, stayontop_bits, true);
  b.setMask(b);
  setBitmap(b);
} // LogikButtonStayOnTop::LogikButtonStayOnTop

void LogikButtonStayOnTop::clickEvent(int)
{
  emit(toggleStayOnTop());
} // LogikButtonStayOnTop::clickEvent
